@@ -1,7 +1,12 @@
 #!/usr/bin/env python
 
-"""Module containing wrapper classes over paramiko.SSHClient
-See SSHClient and ParallelSSHClient
+"""Small wrapper library over paramiko that allows for parallel execution of SSH commands on remote hosts and executing simple single host commands over SSH.
+
+parallel-ssh uses asychronous network requests - there is *no* multi-threading or multi-processing used.
+
+This is a *requirement* for commands on many (hundreds/thousands/hundreds of thousands) of hosts which would grind a system to a halt simply by having so many processes/threads all wanting to execute if done with multi-threading/processing.
+
+See :mod:`pssh.SSHClient` and :mod:`pssh.ParallelSSHClient` for class documentation.
 
 Copyright (C) 2014 Panos Kittenis
 
@@ -19,13 +24,13 @@ License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 """
 
+import logging
 import gevent.pool
 from gevent import socket
-import paramiko
-import os
-import logging
 from gevent import monkey
 monkey.patch_all()
+import paramiko
+import os
 
 host_logger = logging.getLogger('host_logging')
 handler = logging.StreamHandler()
@@ -37,7 +42,6 @@ NUM_RETRIES = 3
 
 logger = logging.getLogger(__name__)
 
-
 def _setup_logger(_logger):
     """Setup default logger"""
     _handler = logging.StreamHandler()
@@ -45,7 +49,6 @@ def _setup_logger(_logger):
     _handler.setFormatter(log_format)
     _logger.addHandler(handler)
     _logger.setLevel(logging.DEBUG)
-
     
 class UnknownHostException(Exception):
     """Raised when a host is unknown (dns failure)"""
@@ -62,6 +65,11 @@ class AuthenticationException(Exception):
     pass
 
 
+class ProxyCommandException(Exception):
+    """Raised on ProxyCommand error - ProxyCommand configured exited with error"""
+    pass
+
+
 class SSHClient(object):
     """Wrapper class over paramiko.SSHClient with sane defaults
     Honours ~/.ssh/config and /etc/ssh/ssh_config entries for host username \
@@ -69,7 +77,8 @@ class SSHClient(object):
 
     def __init__(self, host,
                  user=None, password=None, port=None,
-                 pkey=None):
+                 pkey=None, forward_ssh_agent=True,
+                 _agent=None):
         """Connect to host honouring any user set configuration in ~/.ssh/config \
         or /etc/ssh/ssh_config
          
@@ -86,14 +95,25 @@ class SSHClient(object):
         :type port: int
         :param pkey: (Optional) Client's private key to be used to connect with
         :type pkey: :mod:`paramiko.PKey`
+        :param forward_ssh_agent: (Optional) Turn on SSH agent forwarding - \
+        equivalent to `ssh -A` from the `ssh` command line utility. \
+        Defaults to True if not set.
+        :type forward_ssh_agent: bool
+        :param _agent: (Optional) Override SSH agent object with the provided. \
+        This allows for overriding of the default paramiko behaviour of \
+        connecting to local SSH agent to lookup keys with our own SSH agent. \
+        Only really useful for testing, hence the internal variable prefix.
+        :type _agent: :mod:`paramiko.agent.Agent`
         :raises: :mod:`pssh.AuthenticationException` on authentication error
         :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
-        :raises: :mod:`pssh.ConnectionErrorException` on error connecting"""
+        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
+        :raises: :mod:`pssh.ProxyCommandException` on error with ProxyCommand configured
+        """
         ssh_config = paramiko.SSHConfig()
         _ssh_config_file = os.path.sep.join([os.path.expanduser('~'),
                                              '.ssh',
                                              'config'])
-        # Load ~/.ssh/config if it exists to pick up username
+        # Load ~/.ssh/config if it exists to pick up username, ProxyCommand
         # and host address if set
         if os.path.isfile(_ssh_config_file):
             ssh_config.parse(open(_ssh_config_file))
@@ -108,6 +128,7 @@ class SSHClient(object):
             user = _user
         client = paramiko.SSHClient()
         client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
+        self.forward_ssh_agent = forward_ssh_agent
         self.client = client
         self.channel = None
         self.user = user
@@ -115,6 +136,13 @@ class SSHClient(object):
         self.pkey = pkey
         self.port = port if port else 22
         self.host = resolved_address
+        self.proxy_command = paramiko.ProxyCommand(host_config['proxycommand']) if 'proxycommand' in \
+                             host_config else None
+        if self.proxy_command:
+            logger.debug("Proxy configured for destination host %s - ProxyCommand: '%s'" % (
+                self.host, " ".join(self.proxy_command.cmd),))
+        if _agent:
+            self.client._agent = _agent
         self._connect()
 
     def _connect(self, retries=1):
@@ -122,7 +150,8 @@ class SSHClient(object):
         try:
             self.client.connect(self.host, username=self.user,
                                 password=self.password, port=self.port,
-                                pkey=self.pkey)
+                                pkey=self.pkey,
+                                sock=self.proxy_command)
         except socket.gaierror, e:
             logger.error("Could not resolve host '%s'", self.host,)
             while retries < NUM_RETRIES:
@@ -141,6 +170,10 @@ class SSHClient(object):
                                                                     self.port,))
         except paramiko.AuthenticationException, e:
             raise AuthenticationException(e)
+        except paramiko.ProxyCommandFailure, e:
+            logger.error("Error executing ProxyCommand - %s" % (
+                e.message,))
+            raise ProxyCommandException(e.message)
 
     def exec_command(self, command, sudo=False, **kwargs):
         """Wrapper to :mod:`paramiko.SSHClient.exec_command`
@@ -162,6 +195,8 @@ class SSHClient(object):
         stderr are buffers containing command output.
         """
         channel = self.client.get_transport().open_session()
+        if self.forward_ssh_agent:
+            agent_handler = paramiko.agent.AgentRequestHandler(channel)
         channel.get_pty()
         # stdin (unused), stdout, stderr
         (_, stdout, stderr) = (channel.makefile('wb'), channel.makefile('rb'),
@@ -232,12 +267,13 @@ class ParallelSSHClient(object):
     """Uses :mod:`pssh.SSHClient`, performs tasks over SSH on multiple hosts in \
     parallel.
 
-    Connections to hosts are established in parallel when an object of this
-    class is created, therefor any connection and/or authentication exceptions \
-    will happen on creation and need to be caught."""
+    Connections to hosts are established in parallel when ``exec_command`` is called,
+    therefor any connection and/or authentication exceptions will happen on the
+    call to ``exec_command`` and need to be caught."""
 
     def __init__(self, hosts,
                  user=None, password=None, port=None, pkey=None,
+                 forward_ssh_agent=True,
                  pool_size=10):
         """
         :param hosts: Hosts to connect to
@@ -259,16 +295,17 @@ class ParallelSSHClient(object):
         :raises: :mod:`pssh.AuthenticationException` on authentication error
         :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
         :raises: :mod:`pssh.ConnectionErrorException` on error connecting
+        :raises: :mod:`pssh.ProxyCommandException` on error with ProxyCommand configured
         
         **Example**
 
         >>> from pssh import ParallelSSHClient, AuthenticationException,\
         		UnknownHostException, ConnectionErrorException
+        >>> client = ParallelSSHClient(['myhost1', 'myhost2'])
         >>> try:
-        >>> ... client = ParallelSSHClient(['myhost1', 'myhost2'])
+        >>> ... cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
         >>> except (AuthenticationException, UnknownHostException, ConnectionErrorException):
         >>> ... return
-        >>> cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
         >>> output = [client.get_stdout(cmd) for cmd in cmds]
         [myhost1]     ls: cannot access /tmp/aasdfasdf: No such file or directory
         [myhost2]     ls: cannot access /tmp/aasdfasdf: No such file or directory
@@ -279,11 +316,11 @@ class ParallelSSHClient(object):
 
         >>> from pssh import ParallelSSHClient, AuthenticationException,\
         		UnknownHostException, ConnectionErrorException
+        >>> client = ParallelSSHClient(['myhost1', 'myhost2'])
         >>> try:
-        >>> ... client = ParallelSSHClient(['myhost1', 'myhost2'])
+        >>> ... cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
         >>> except (AuthenticationException, UnknownHostException, ConnectionErrorException):
         >>> ... return
-        >>> cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
         >>> output = [client.get_stdout(cmd, return_buffers=True) for cmd in cmds]
         >>> print output
         [{'myhost1': {'exit_code': 2,
@@ -330,6 +367,7 @@ class ParallelSSHClient(object):
         self.hosts = hosts
         self.user = user
         self.password = password
+        self.forward_ssh_agent = forward_ssh_agent
         self.port = port
         self.pkey = pkey
         # To hold host clients
@@ -371,7 +409,8 @@ class ParallelSSHClient(object):
         if not self.host_clients[host]:
             self.host_clients[host] = SSHClient(host, user=self.user,
                                                 password=self.password,
-                                                port=self.port, pkey=self.pkey)
+                                                port=self.port, pkey=self.pkey,
+                                                forward_ssh_agent=self.forward_ssh_agent)
         return self.host_clients[host].exec_command(*args, **kwargs)
 
     def get_stdout(self, greenlet, return_buffers=False):
@@ -444,7 +483,8 @@ class ParallelSSHClient(object):
         if not self.host_clients[host]:
             self.host_clients[host] = SSHClient(host, user=self.user,
                                                 password=self.password,
-                                                port=self.port)
+                                                port=self.port, pkey=self.pkey,
+                                                forward_ssh_agent=self.forward_ssh_agent)
         return self.host_clients[host].copy_file(local_file, remote_file)
 
     
